//
//  ClusterBuilder.cpp
//
//  Copyright (c) Tomoya Mori. All rights reserved.
//


#include "ClusterBuilder.h"

const string ClusterBuilder:: summary = "summary_";
const string ClusterBuilder:: map_data = "cluster_map_";
const string ClusterBuilder:: core_data = "cluster_core_";
const string ClusterBuilder:: event_log = "event_log_";
const string ClusterBuilder:: comp_log = "log_";

//const string ClusterBuilder:: criterionName[_criterion_end] =
//{
//    "AIC",
//    "BIC"
////    ,
////    "NML"
//};

const string ClusterBuilder:: eventName[_event_end] =
{
    "FORMATION",
    "EXPANSION"
//    ,
//    "DELETION"
};

ClusterBuilder::
ClusterBuilder(// int in_criterion, 
               set<Region*>* in_all_regions,
               double in_expand_step,
               double in_national_area,
               double in_industry_count,
               Distance* in_dist)
//               , map<int, vector<double>* >* in_penalty_table )
{
    all_regions = in_all_regions;
    expand_search_step = in_expand_step;
    total_area = in_national_area;
    total_count = in_industry_count;
    distance = in_dist;
    log_count = 0;
    if_random = false;
//    criterion_choice = in_criterion;
    
//    if ( criterion_choice == 2 )
//        penalty_table = in_penalty_table;
}

double ClusterBuilder::
Evaluate( ClusterSystem* in_system )
{
    return in_system->Evaluate();
}

Cluster* ClusterBuilder::
GenerateFirstCluster( Region* initial_region )
{
    current_system = new ClusterSystem( total_area,
                                       total_count,
                                       *all_regions );
    
    Cluster* new_cluster = new Cluster( initial_region, 
                                       distance );
    
    current_system->AddMember( new_cluster );
    current_system->CalcTc();
    
    return new_cluster;
    
}

Cluster* ClusterBuilder::
GenerateFirstCluster()
{
    current_system = new ClusterSystem( total_area,
                                       total_count,
                                       *all_regions );
    // Construct cluster scheme with no clusters.
    
    Region* best_cluster;
    
    double prev_tc = -1.0;
    double current_tc;
    
    for ( set<Region*>::iterator i1 =
         all_regions->begin(); i1 != all_regions->end(); ++i1 ) 
    {
        if ( (**i1).Count() < 0.5 )
            continue;
        // Search regions with positive counts.
        
        double l_hat_c = 
        (**i1).Count() * log( (**i1).Count() / total_count );
        
        double l_0 = 
        (**i1).Count() * log( (**i1).Area() / total_area );
        
        double residual_count = total_count - (**i1).Count();
        double residual_area = total_area - (**i1).Area();
        
        if ( residual_count > 0.0 )
        {
            l_hat_c += residual_count * log ( residual_count / total_count );
            l_0 += residual_count * log ( residual_area / total_area );
        }
        
        current_tc = l_hat_c - l_0;
        
        if ( current_tc > prev_tc )
        {
            best_cluster = *i1;
            prev_tc = current_tc;
        }
    }
    
    Cluster* new_cluster = new Cluster( best_cluster, distance );
    current_system->AddMember( new_cluster );
    current_system->CalcTc();
    
    return new_cluster;
}


Cluster* ClusterBuilder::
GenerateNewCluster( Region* target,
                   double& new_value,
                   int& new_location )
{
    Cluster* new_cluster = new Cluster( target,
                                       distance );
    
    potential_system->AddMember( new_cluster );
    potential_system->CalcTc();
    new_value = Evaluate( potential_system );
    new_location = new_cluster->Location();
    return new_cluster;
}

void  ClusterBuilder::
GenerateNewCluster( double& new_value,
                   int& new_location )
{
    candidate_new_cluster = new ClusterSystem( *current_system );
    // best cluster system with a new cluster
    
    potential_system = new ClusterSystem( *current_system );
    // cluster system with a new cluster
    
    new_value = 0.0;
    // best criterion value

    
    /// Identify potential regions to construct a new cluster ///
    
    set<Region*> potential_locations;
    
    for ( set<Region*>::iterator i1 = current_system->ResidualRegions().begin();
         i1 != current_system->ResidualRegions().end(); ++i1 )
    {
        if ( (**i1).Count() < 0.5 )
            continue;
        
        potential_locations.insert( *i1 );
    }
    
    if ( potential_locations.size() == 0 )
    {
        new_value = -1.0;
        new_location = -1.0;
        return;
    }
    
    
    /// Identify the best local improvement of cluster scheme.
    
    for ( set<Region*>::iterator i1 = potential_locations.begin();
         i1 != potential_locations.end(); ++i1 )
    {
        double trial_value;
        int trial_cluster;
        Cluster* new_cluster = GenerateNewCluster( *i1, 
                                                  trial_value,
                                                  trial_cluster );
        // Add new_cluster to potential_system.
        
        if ( trial_value <= new_value )
        {
            delete potential_system;
            potential_system = new ClusterSystem( *current_system );
        }
        else
        {
            // Update the best improvement.

            delete candidate_new_cluster;
            
            new_value = trial_value;
            new_location = new_cluster->Location();
            
            candidate_new_cluster = 
            new ClusterSystem( *potential_system );
            
            delete potential_system;	  
            potential_system = new ClusterSystem( *current_system );
        }
    }
    delete potential_system;
}


double  ClusterBuilder::
ExpandCluster( double& new_value,
              int& expanded_location,
              int& consolidated_location )
{
    // Set the best system as current system
    candidate_expansion = new ClusterSystem( *current_system );
    
    // Set the default of potential system as current system
    potential_system = new ClusterSystem( *current_system );
    
    //  new_value = current_system->Criterion();
    new_value = -1.0;
    
    // Location which expansion takes place :
    expanded_location = 0;
    consolidated_location = 0;
    double dist_to_consolidated_location = 0.0;
    
    for ( set<Cluster*>::iterator i1 = current_system->Members().begin();
         i1 != current_system->Members().end(); ++i1 )
    {
        // Target cluster to expand
        Cluster* target = NULL;
        
        int tmp_consolidated_loc;
        double tmp_dist;
        
        
        /// Order sig regions by the proximity to target 
        
        list<Region*> neighbor_regions;
        
        (**i1).FindNeighborRegions( *all_regions,
                                   neighbor_regions,
                                   expand_search_step );
        // Identifies the set of basic regions within a "expand_search_step" neighborhood of
        // cluster i1 (toward which this cluster is to be expanded).
        
        if ( neighbor_regions.size() == 0 )
        {
            // There is no basic regions to expand within a given search range (expand_serch_step).
            continue;
        }
        
        double trial_value = 0.0;
        // To store a criterion value for each potential expansion.
        
        set<int> tested_clusters;
        // The set of clusters for which possible expansions have been tested.
        
        bool if_tested_clusters;
        // True if included in tested_clusters.
                
        for ( list<Region*>::iterator i2 = neighbor_regions.begin(); 
             i2 != neighbor_regions.end(); ++i2 )
        {
            if_tested_clusters = false;
            
            /// Identify target cluster in the potential_system ///
            // target cluster (i1) in current_system must be identified in potential_sytem
            // at each iteration of neighbor region i2, since potential_system may be updated for
            // each i2.
            
            for ( set<Cluster*>::iterator i3 = 
                 potential_system->Members().begin();
                 i3 != potential_system->Members().end(); ++i3 )
            {
                // Identify the cluster in potential_system corresponding to i1.
                
                if ( (**i1) == (**i3 ) )
                {
                    target = *i3;
                    break;
                }
            }
            
            bool if_member = false;
            // True if a region in the neighbor 
            // belongs to a cluster.
            
            Cluster* cluster_tobe_merged;
            
            for ( set<Cluster*>::iterator i3 =
                 potential_system->Members().begin();
                 i3 != potential_system->Members().end(); ++i3 )
            {
                if ( (**i3) == *target )
                    continue;
                
                if ( (**i3).IfCoreMember( *i2 ) )
                {
                    // Basic region i2 belongs to an existing cluster i3.
                    
                    if ( tested_clusters.find( (**i3).Location() ) != 
                        tested_clusters.end()  )
                    {
                        // The potential expansion of cluster i3 has
                        // already been tested.
                        
                        if_tested_clusters = true;
                        
                        break;
                    }
                    else
                        tested_clusters.insert((**i3).Location());
                    
                    if_member = true;
                    cluster_tobe_merged = *i3;
                    break;
                }
            }
            
            
            if ( if_tested_clusters )
                continue;
            // Skip if the merger to this cluster has been already tested.
            
            // Expand target cluster :
            
            if ( if_member ) 
            {
                tmp_consolidated_loc = cluster_tobe_merged->Location();
                tmp_dist = target->DistanceTo( cluster_tobe_merged );
                target->Expand( cluster_tobe_merged );
                // Merge cluster_tobe_merged to target (but not D-convex solidify yet).
            }
            else
            {
                tmp_consolidated_loc = (**i2).Location();
                tmp_dist = target->DistanceTo( *i2 );
                target->Expand( *i2 );
                // Merge basic region i2 to target (but not D-conves solidify yet).
            }
            
            
            set<Cluster*> delete_clusters;	  
            
            target->FindConvergentClosure( *all_regions, 
                                          potential_system->Members(),
                                          delete_clusters );
            // D-convex solidfy target after merger.
            
            potential_system->ShrinkResidualRegions( target );
            // Update residual regions.
            
            potential_system->EraseMember( delete_clusters );
            // Delete merged clusters.
            
            potential_system->CalcTc();
            // Compute loglikelihood of potential_system
            
            trial_value = Evaluate( potential_system );
            // Compute BIC value and save it to trial_value
            
            if ( trial_value <= new_value )
            {
                // Expansion of cluster i1 toward i2 does not improve BIC.
                delete potential_system;
                potential_system = new ClusterSystem( *current_system );
            }
            else
            {
                // Expansion of cluster i1 toward i2 improves BIC
                expanded_location = target->Location();
                consolidated_location = tmp_consolidated_loc;
                dist_to_consolidated_location = tmp_dist;
                new_value = trial_value;
                
                delete candidate_expansion;
                candidate_expansion = 
                new ClusterSystem( *potential_system );
                
                delete potential_system;
                potential_system = new ClusterSystem( *current_system );
            }
        }
    } 
    
    
    
    delete potential_system;
    
    if ( new_value < 0.0 )
    {
        return -1.0;
    }
    
    return dist_to_consolidated_location;
}




//double ClusterBuilder::
//Delete( Cluster* target )
//{
//    potential_system->EraseMember(target);
//    potential_system->CalcTc();
//    return Evaluate( potential_system );
//}
//
//int ClusterBuilder::
//Delete(	double& new_value,
//       int& deleted_location )
//{
//    // Set the best system as current system
//    candidate_expansion = new ClusterSystem( *current_system );
//    
//    // Set the default of potential system as current system
//    potential_system = new ClusterSystem( *current_system );
//    
//    //  new_value = current_system->Criterion();
//    
//    for ( set<Cluster*>::iterator i1 = current_system->Members().begin();
//         i1 != current_system->Members().end(); ++i1 )
//    {
//        // Target cluster to expand
//        Cluster* target = NULL;
//        
//        // Find the target cluster in potential system
//        for ( set<Cluster*>::iterator i2 = potential_system->Members().begin();
//             i2 != potential_system->Members().end(); ++i2 )
//        {
//            if ( (**i1) == (**i2 ) )
//            {
//                target = *i2;
//                break;
//            }
//        }
//        
//        int trial_location = target->Location();
//        double trial_value = Delete( target );
//        
//        
//        if ( trial_value <= new_value )
//        {
//            delete potential_system;
//            potential_system = new ClusterSystem( *current_system );
//        }
//        else
//        {
//            deleted_location = trial_location;
//            new_value = trial_value;
//            
//            delete candidate_expansion;
//            candidate_expansion = 
//            new ClusterSystem( *potential_system );
//            
//            delete potential_system;
//            potential_system = new ClusterSystem( *current_system );
//        }
//    }
//    
//    if ( candidate_expansion->Criterion() >
//        current_system->Criterion() )
//    {
//        delete current_system;
//        current_system = new ClusterSystem( *candidate_expansion );
//        
//    }
//    delete candidate_expansion;  
//    return deleted_location;
//}


void ClusterBuilder::
SetOS( const string& out_dir_name )
{
    out_dir = out_dir_name;
}

void ClusterBuilder::
SetEventLog()
{
    string tmp_str;
    if (!if_random)
        eventlog = out_dir + event_log + "ind_" + 
        Conversion( tmp_str, id, 0 ) + ".txt";
    else
        eventlog = out_dir + event_log + "ind_" + 
        Conversion( tmp_str, id, 0 ) + "_" +
        Conversion(tmp_str,sample_id,0) + ".txt";
    
    
    ofstream os( eventlog.c_str() );
//    if (!os) NoFile nf( eventlog );


    
    os << Conversion(tmp_str, "EVENT" );
    os << Conversion(tmp_str, "CLUSTER" );
    os << Conversion(tmp_str, "CONSOLIDATED" );
    os << Conversion(tmp_str, "LIKELIHOOD" );
    os << Conversion(tmp_str, "BIC" ) << endl;
    
    os.close();
    os.clear();
};

void ClusterBuilder::
SetCompLog()
{
    string tmp_str;
    string complog;
    
    if (!if_random)
        complog = out_dir + comp_log + "ind_" + 
        Conversion( tmp_str, id, 0 ) + ".txt";
    else
        complog = out_dir + comp_log + "ind_" + 
        Conversion( tmp_str, id, 0 ) + "_" +
        Conversion(tmp_str,sample_id,0) + ".txt";
    
    
    log_os.open( complog.c_str() );
};


void ClusterBuilder::
GrowSystem( Region* initial_region )
{
    // Criterion values :
    double c_current, c_formation, c_expansion;
    // c_current = present criterion value
    // c_formation = criterion value under new cluster formation
    // c_expansion = criterion value under cluster expansion
    
    // Candidate location of cluster formation/expansion :
    int event_expansion, event_formation, consolidated_location;
    
    
    
    /// Generate first cluster ///
    
    ++log_count;
    
    Cluster* first_cluster;
    
    if ( initial_region != NULL )
    {
        // If the first cluster is specified :
        first_cluster = GenerateFirstCluster(initial_region);
    }
    else
    {
        // Find the best single-region cluster (in terms of BIC)
        first_cluster = GenerateFirstCluster();
    }
    
    c_current = Evaluate( current_system );
    // Compute BIC for the current_system.
    
    if ( c_current < 0.0 )
    {
        // There is no single-region cluster which has likelihood 
        // larger than no-cluster model.
        
        if ( IfTakeLog( _main ) )
        {
            log_os << "\n\nLog : " << log_count << endl;
            log_os << "No cluster is detected." << endl;
            
            log_os << "Current value : " << 0.0 << endl << endl;
        }
        
        delete first_cluster;
        current_system->Members().clear();
        
        
        WriteLog(true);
        return;
        
    }
    
    /// Log for the first cluster construction ///
    
    if ( IfTakeLog( _main ) )
    {
        log_os << "\n\nLog : " << log_count << endl;
        log_os << "First cluster at " 
               << first_cluster->Location()
               << endl;
        log_os << "Current value : " 
               << c_current << endl << endl;
    }
    
    if ( IfTakeLog( _event ) )
        WriteFormationLog(// (int) _formation, 
                          first_cluster->Location() );
    
    if ( IfTakeLog( _map ) )
        WriteLog();
    
    
    /// Construction of the second cluster and more ///
    
    while(true)
    {
        ++log_count;
        
        c_current = Evaluate( current_system);
        // Compute BIC for the current_system.
        
        if ( IfTakeLog( _main ) )
        {
            log_os << "Log : " << log_count << endl;
            log_os << "Current value : " << c_current << endl;
        }
        
        // Find the best new formation of a cluster :
        
        GenerateNewCluster(  c_formation,
                           event_formation );
        
        // A new cluster formation is considered only at
        // the basic regions in the residual region which have positive counts
        
        if ( IfTakeLog( _main ) )
        {
            if ( c_formation == 0.0 )
                log_os << "No possible new formation." << endl;
            else  
                log_os << "Candidate formation at " << event_formation << "(" 
                << c_formation << ")" << endl;
        }
        
        // Find the best expansion of existing clusters :
        
        double dist_to_consolidated_loc =
        ExpandCluster( c_expansion,
                      event_expansion,
                      consolidated_location );
        
        if ( IfTakeLog( _main ) )
        {
            if ( event_expansion > 0 )
            {
                log_os << "Candidate expansion at : " << event_expansion 
                << " to " << consolidated_location 
                << " with dist " 
                << dist_to_consolidated_loc
                << " ("
                << c_expansion << ")" << endl;
            }
            else if ( event_expansion == 0 )
            {
                log_os << "No candidate expansion." << endl;
            }
        }
        
        if ( ( c_current >= c_formation && 
              c_current >= c_expansion ) )
        {
            if ( IfTakeLog( _map ) ) 
            {
                WriteLog();
            }
        }
        else
        {
            delete current_system;
            
            if ( c_formation > c_expansion )
            {
                current_system = new ClusterSystem( *candidate_new_cluster );
                // The cluster scheme with a new cluster has been stored in 
                // candiate_new_cluster.
                
                if ( IfTakeLog(_event) )
                    WriteFormationLog(// _formation, 
                                      event_formation );
                
                if ( IfTakeLog(_map) )
                    WriteLog();
                
                if ( IfTakeLog( _main ) )
                    log_os << "Formation : Num clusters = " << flush;
            }
            else
            {
                current_system = new ClusterSystem( *candidate_expansion );
                // THe cluster scheme with cluster expansion has been stored in
                // candidate_expansion.
                
                if ( IfTakeLog(_event) )
                    WriteExpansionLog(// _expansion, 
                                   event_expansion, 
                                   consolidated_location );
                
                if ( IfTakeLog(_map) )
                    WriteLog();
                
                if ( IfTakeLog(_main) )
                    log_os << "Expansion : Num clusters = " << flush;
            }
            
            if ( IfTakeLog(_main) )
                log_os << current_system->Members().size() 
                << endl << endl;
        }
        
        delete candidate_new_cluster;
        delete candidate_expansion;
        
        if ( ( c_current >= c_formation && 
              c_current >= c_expansion ) )
        {
            // Cluster detection complete :
            // No expansion/deletion of the existing clusters, nor formation of 
            // new clusters improve the system.
            WriteLog(true);
            break;
        }
    }
    
    if ( IfTakeLog(_main) )
        log_os << "Cluster system has been found ...\n" << endl;
}


void ClusterBuilder::
WriteFormationLog( //int in_event,
              int in_cluster )
{
    string tmp_str;
    ofstream os( eventlog.c_str(), ios::out|ios::app  );
//    if (!os) NoFile nf( eventlog );

    
    os << Conversion( tmp_str, eventName[_formation] );
    os << Conversion( tmp_str, in_cluster );
    os << Conversion( tmp_str, "_" );
    os << Conversion( tmp_str, current_system->Tc() );
    os << Conversion( tmp_str, current_system->Criterion()  ) << endl;
    
    os.close();
    os.clear();
}

void ClusterBuilder::
WriteExpansionLog( //int in_event,
               int in_cluster,
               int in_included )
{
    string tmp_str;
    ofstream os( eventlog.c_str(), ios::out|ios::app  );
//    if (!os) NoFile nf( eventlog );

    
    os << Conversion( tmp_str, eventName[_expansion] );
    os << Conversion( tmp_str, in_cluster );
    os << Conversion( tmp_str, in_included);
    os << Conversion( tmp_str, current_system->Tc() );
    os << Conversion( tmp_str, current_system->Criterion()) << endl;
    
    os.close();
    os.clear();
}

void ClusterBuilder::
WriteLog( bool if_final )
{
    // Summary //
    string tmp_str, outname;
    
    if ( !if_final )
    {
        outname = out_dir + summary + "ind_" + 
        Conversion(tmp_str,id,0) + "_log_" +
        Conversion(tmp_str,log_count,0 ) + ".csv";
    }
    else
    {
        if ( !if_random )
        {
            outname = out_dir + summary + "ind_" + 
            Conversion(tmp_str,id,0) + ".csv";
        }
        else
        {
            outname = out_dir + summary + "ind_" + 
            Conversion(tmp_str,id,0) + "_" +
            Conversion(tmp_str,sample_id,0) + ".csv";
        }
    }
    
    
    ofstream os( outname.c_str() );
//    if (!os) NoFile nf( outname );

    current_system->Update();
    current_system->WriteSummary( os, *all_regions );
    
    // Convex-solid area of clusters //
    
    if ( !if_final )
    {
        outname = out_dir + map_data + "ind_" +  
        Conversion(tmp_str,id,0) + "_log_" +
        Conversion(tmp_str,log_count,0 ) + ".csv";
    }
    else
    {
        if ( !if_random )
        {
            outname = out_dir + map_data + "ind_" +  
            Conversion(tmp_str,id,0) + ".csv";
        }
        else
        {
            outname = out_dir + map_data + "ind_" +  
            Conversion(tmp_str,id,0) + "_" +
            Conversion(tmp_str,sample_id,0) + ".csv";
        }
    }
    os.open( outname.c_str() );
    current_system->WriteMapData( os );
    
    // "Core" region of clusters //
    
    if ( !if_final )
    {
        outname = out_dir + core_data + "ind_" +  
        Conversion(tmp_str,id,0) + "_log_" +
        Conversion(tmp_str,log_count,0 ) + ".csv";
    }
    else
    {
        if ( !if_random )
        {
            outname = out_dir + core_data + "ind_" +  
            Conversion(tmp_str,id,0) + ".csv";
        }
        else
        {
            outname = out_dir + core_data + "ind_" +  
            Conversion(tmp_str,id,0) + "_" +
            Conversion(tmp_str,sample_id,0) + ".csv";
        }
    }
    
    os.open( outname.c_str() );
    current_system->WriteCore( os );
    
}

ClusterBuilder::
~ClusterBuilder()
{
    if ( current_system != NULL )
        delete current_system;
}



