//
//  Distance.cpp
//
//  Copyright (c) Tomoya Mori. All rights reserved.
//

#include "Distance.h"

double Distance::
operator()( int orig, int dest ) 
{
    if ( orig == dest )
        return 0.0;
    else
    { 
        if ( orig > dest )
        {
            int tmp_dest = orig;
            orig = dest;
            dest = tmp_dest;
        }
        
        if ( data.find(orig) == data.end() )
        {
            string tmp_str;
            string error_message;
            //	    cout << "No distance data between " << orig << " and " << dest << endl;
            error_message.append("\nERROR: DISTANCE::operator() No data for ORIGIN ");
            error_message.append( Conversion( tmp_str, orig, 0 ) );
            error_message.append(" for DESTINATION ");
            error_message.append( Conversion(tmp_str, dest,0) );
            error_message.append( "\n" );
            throw(error_message);
        }
        else if ( (*(data[orig])).find( dest ) == (*(data[orig])).end() )
        {	
            return DBL_MAX;
        }
        
        return(*(data[orig]))[ dest ];
    }
}

Distance& Distance::
operator=( const Distance& in_dist )
{
    for ( map<int, map<int,double>* >::const_iterator i = in_dist.Data().begin();
         i != in_dist.Data().end(); ++i )
    {
        map<int, double>* new_data = new map<int,double>;
        data[i->first] = new_data;
        
        for ( map<int,double>::const_iterator j = i->second->begin();
             j != i->second->end(); ++j )
        {
            new_data->insert( *j );
        }
    }
    return *this;
}


void Distance::
Read( istream& is )
{
    Csv1 line;
    int num = 0;
    cout << "\nReading distance data ... " << flush;
    while ( is.peek() != EOF )
    {
        is >> line;
        
        if ( line.Label() == "INDEX" ) continue;
        
        int orig, dest;
        double val;
        Conversion(orig,line[0]);
        Conversion(dest,line[1]);
        Conversion(val,line[2]);
        
        if ( orig > dest ) continue;
        
        if ( data.find(orig) == data.end() )
        {
            map<int,double>* new_data = new map<int,double>;
            data[orig] = new_data;
        }
        
        if ( data.find(dest) == data.end() )
        {
            map<int,double>* new_data = new map<int,double>;
            data[dest] = new_data;
        }
        
        data[orig]->insert( pair<int,double>(dest,val) );
        ++num;
        if ( num % 100000 == 0 )
            cout << "." << flush;
    }
    cout << endl;
}

//void Distance::
//Write( ostream& os )
//{
//    string tmp_str;
//    
//    for ( map< int, map<int,double>* >::iterator i = data.begin();
//         i != data.end(); ++i )
//    {
//        
//        for ( map<int,double>::iterator j = i->second->begin();
//             j != i->second->end(); ++j )
//        {
//            
//            os << Conversion( tmp_str,Conversion(tmp_str, (*i).first,0) + "," )
//            << Conversion( tmp_str, Conversion(tmp_str, (*j).first,0) + "," )
//            << Conversion( tmp_str, Conversion(tmp_str, (*this)((*i).first, (*j).first),0) + "," )
//            << endl;
//        }
//    }
//}

Distance::
~Distance()
{
    for ( map<int, map<int,double>* >::iterator i = data.begin();
         i != data.end(); ++i )
    {
        delete i->second;
    }
    
}


