//
//  NamedObjectVector.h
//
//  Copyright (c) Tomoya Mori. All rights reserved.
//

#ifndef DetectClusters_NamedObjectVector_h
#define DetectClusters_NamedObjectVector_h

#include "StringManipulator.h"
#include "IOAdapter.h"
#include <string>
#include <vector>

using namespace std;

template< class T >
class NamedObjectVector : public IOAdapter, public vector< T >
{
private:
    
    static const string default_label;
    
protected:
    
    ////////// Data and methods for initialization //////////
    
    string label;
    
    const string&
    DefaultLabel() const { return default_label; }
    
public:
    
    ////////// Exceptions /////////
    
    class BadIndex
    {
    public:
        int badindex;
        BadIndex(int i) : badindex(i) {}
    };
    // Thrown when non-existent i-th element is called.
    
    class BadLabel
    {
    public:
        string badlabel;
        BadLabel(string s) : badlabel(s) {}
    };
    // Thrown if label with more than 
    // one ws-separated strings is requested.
    
    
    ////////// Methods for initialization //////////
    
    void 
    Label( const string& inLabel ) throw(BadLabel&);
    
    const string&
    Label() const { return label; }
    
    ////////// Constructors/Destructor //////////
    
    NamedObjectVector( const string inLabel = "" ) { Label( inLabel ); };
    NamedObjectVector( const NamedObjectVector<T>& inNamedObjectVector ) 
	{ operator=( inNamedObjectVector ); }
    NamedObjectVector( const vector<T>& inVector, const string inLabel = "" )
	{ 
	    Label( inLabel );
	    operator=( inVector );
	}
    
    virtual ~NamedObjectVector() {}
    
    
    ////////// Operators //////////
    
    virtual NamedObjectVector&
    operator=( const NamedObjectVector& inNamedObjectVector );
    
    virtual NamedObjectVector& 
    operator=( const vector< T >& inVector );
    
    T& operator[](int i) throw(BadIndex &);
    
    const T& operator[](int i) const throw(BadIndex &);
};

////////// Implementation //////////

template< class T >
const string NamedObjectVector< T > :: 
default_label = "..........";

template< class T >
void NamedObjectVector< T > :: 
Label( const string& inLabel ) throw(BadLabel&)
{ 
    string label1, label2, source = inLabel;
    string delimiter = ", \t\n";
    label1 = Token(source,delimiter);
    label2 = Token(source,delimiter);
    
    if ( !label2.empty() )
        throw BadLabel(inLabel);
    else
        label = label1; 
}

template< class T > 
NamedObjectVector< T >& NamedObjectVector< T > :: 
operator=( const NamedObjectVector< T >& inNamedObjectVector )
{
    if ( this == &inNamedObjectVector )
        return *this;
    
    Label( inNamedObjectVector.Label() );
    vector<T>::operator=( inNamedObjectVector );
    
    return *this;
}

template< class T > 
NamedObjectVector< T >& NamedObjectVector< T > :: 
operator=( const vector< T >& inVector )
{
    vector< T >::operator=( inVector );
    return *this;
}

template< class T >
T& NamedObjectVector< T > :: 
operator[](int i) throw(BadIndex &)
{
    if ( i < 0 || i > this->size()-1 )
        throw BadIndex(i);
    
    return vector<T>::operator[](i);
}

template< class T >
const T& NamedObjectVector< T > :: 
operator[](int i) const throw(BadIndex &)
{
    if ( i < 0 || i > this->size()-1 )
        throw BadIndex(i);
    
    return vector<T>::operator[](i);
}



#endif
