//
//  StringManipulator.cpp
//
//  Copyright (c) Tomoya Mori. All rights reserved.
//

#include "StringManipulator.h"

void
Separate( const string& buf,
         string& str1,
         string& str2,
         const string delimiter, 
         const string separator )
{
    str1.erase();
    str2.erase();
    
    int str1_end = (int) buf.find( delimiter );
    
    if ( str1_end == string::npos ) 
        // no delimieter, i.e., str2 does not exist
        str1 = buf;
    
    else
    {
        // if ( buf[str1_end] != buf[0] ) 020501
        if ( str1_end != 0 ) // str1 exists
            str1.assign( buf.begin(), buf.begin() + str1_end );
        
        unsigned long source_str2_begin = str1_end + delimiter.length();
        
        if ( source_str2_begin < buf.size() ) // str2 exists
        {
            string source_str2;
            source_str2.assign( buf.begin()+source_str2_begin, buf.end() );
            str2 = Token( source_str2, separator );
        }
    }
}


string
Token( string& source,
      const string delimiter )
{
    string token;
    int token_first, token_last;
    token_first = (int) source.find_first_not_of( delimiter );
    // first non-delimiter pos in source
    
    if ( token_first != string :: npos ) // Value exists
    { 
        token_last = (int) source.find_first_of( delimiter, token_first );
        // first delimiter pos in source
        
        if ( token_last != string :: npos )  // Source remains
        {
            token.assign( source, token_first, token_last - token_first);
            source.assign( source, token_last, source.size() - token_last );
            // pop first token from source
        }
        
        else // No source remains
        {
            token.assign( source, token_first, source.size() - token_first );
            source.erase();
        }
    }
    
    // If source contains no value, token is empty.
    
    return token;
}

double&
Conversion( double& out,
           const string in )
{
    out = atof( in.c_str() );
    return out;
}


int&
Conversion( int& out,
           const string in )
{
    out = atoi( in.c_str() );
    return out;
}

string&
Conversion( string &out,
           const string in,
           int length_data )
{
    out.erase();
    
    
    int x = (int) in.find_first_not_of( " " ); 
    string in_wo_ws = string( in, x, in.size() - x);
    
    
    if ( ( length_data + 1 - (int) in_wo_ws.size() ) > 0 )
    {
        string ws( length_data - in_wo_ws.size(), ' ');
        out = ws + in_wo_ws;
    }
    else
    {
        out = in_wo_ws;
    }
    
    
    //    if ( ( length_data + 1 - (int) in.size() ) > 0 )
    //      {
    //        string ws( length_data - in.size(), ' ');
    //        out = ws + in;
    //      }
    //    else
    //      {
    //  //      out = in;
    //  	int x = in.find_first_not_of( " " );
    //  	out = string( in, x, in.size() - x);
    //      }
    
    return out;
}

const string 
CommentOut( const string& in,
           int length_data )
{
    string out;
    
    if ( ( length_data + 1 - in.size() ) > 0 )
    {
        string ws( length_data - in.size()-1, ' ');
        out = "#" + ws + in;
    }
    else
        out = "#" + in;
    
    return out;
}

string 
CommentOut( string& in,
           int length_data )
{
    string out;
    
    if ( ( length_data + 1 - in.size() ) > 0 )
    {
        string ws( length_data - in.size()-1, ' ');
        out = "#" + ws + in;
    }
    else
        out = "#" + in;
    
    return out;
}

string&
Conversion( string &out,
           const int &in,
           int length_data,
           int length_decimal )
{
    ostringstream outstr;
    //  outstr.setf(ios::scientific); 020501
    //  outstr.setw(length_data);
    // outstr.width(length_data);
    // column width for "out" is at least length_data
    // (left aligned by default)
    outstr.width(length_data);
    outstr.precision( length_decimal );
    //  outstr << setw(length_data) << in;
    outstr << in;
    out = outstr.str();
    return out;
    //    Conversion( out, outstr.str() );
}


string &
Conversion( string &out,
           const double &in,
           int length_data,
           int length_decimal )
{
    ostringstream outstr;
    //  outstr.setf(ios::scientific);
    outstr.width(length_data);
    outstr.precision( length_decimal );
    
    outstr << in;
    
    out = outstr.str();
    
    return out;
}




