//
//  StringManipulator.h
//
//  Copyright (c) Tomoya Mori. All rights reserved.
//

#ifndef DetectClusters_StringManipulator_h
#define DetectClusters_StringManipulator_h

#include <stdlib.h> // For conversion between char and int/double
#include <iomanip>  //  For I/O format
#include <sstream>
#include <iostream>
#include <string>
#include <cstdio>

using namespace std;

const int LENGTH_DATA = 20;
const int LENGTH_DECIMAL = 7;

void 
Separate( const string& buf,
         string& substr1,
         string& substr2,
         const string delimiter,
         const string separator=", \n\0\t" ); 
// Copy left (right) of delimitor in buf to substr1 (substr2).
// (sbustr2 contains only the first token separated by separator.)

string 
Token( string& source,
      const string delimiter);   
// Returns a string segment upto the first delimiter in source.

//
// In the Conversion() functions below,
// length_data = 0  means free length.
//


////////// Data Type Conversion Functions /////////

double&
Conversion( double& out, const string in );
// Convert a string to a double.

int&
Conversion( int& out, const string in );
// Convert a string to an integer.

string&
Conversion( string& out, const int& in, 
           int length_data = LENGTH_DATA,
           int length_decimal = LENGTH_DECIMAL );
// Convert an integer to a string.
// Column width at least length_data (left aligned)
// Best of scientifc and fixed expressions
// Decimal presition = length_decimal for scientific expression

string&
Conversion( string& out, const double& in,
           int length_data = LENGTH_DATA,
           int length_decimal = LENGTH_DECIMAL );
// Convert a double to a string.
// See Conversion(string&,const int&,int,int)

string&
Conversion( string& out, const string in,
           int length_data = LENGTH_DATA );
// Left align a string "in" with column width = length_data.
// No conversion if size of "in" exceeds length_data.



///////// GnuPlot Comment /////////

const string
CommentOut( const string& in,
           int length_data = LENGTH_DATA );

string
CommentOut( string& in,
           int length_data = LENGTH_DATA );
// Put # at the head of in-string.
// If length_data > in-string.size(), white space is inserted btw
// # and in-string.


#endif
