//
//  main.cpp
//
//  Copyright (c) Tomoya Mori. All rights reserved.
//


#include <fstream>
#include <utility>
#include <string>
#include <vector>
#include <map>
#include <set>
#include "Csv1.h"
#include "Distance.h"
#include "Region.h"
#include "Cluster.h"
#include "ClusterSystem.h"
#include "ClusterBuilder.h"
#include "IndustryLocationData.h"
#include "ShortestPathData.h"
#include "NoFile.h"
#include "CantOpenFile.h"
#include "StringManipulator.h"

using namespace std;
typedef map<int, map<string, double>* > loc_source_data;

int main()
{
    try 
    {
        cout << "Detecting clusters ..." << endl;
        cout << "\nINPUT  : input_detectclusters.txt" 
        << endl;
        
        ofstream os;
        string tmp_str;
        double tmp_dbl;
        int tmp_int;
        
        ifstream is( "./input_detectclusters.txt" );
        Csv1 line;
        
        ///// Target industries /////
        
        is >> line;
        int start_id;
        Conversion( start_id, line[0] );
        
        is >> line;
        int end_id;
        Conversion( end_id, line[0] );
        
        ///// Read industry list /////
        is >> line;
        string ind_list_file = line[0];
        
        
        ///// Criterion choice /////
//        is >> line;
//        int criterion_choice;
//        Conversion( criterion_choice, line[0] );
//        
        ///// Delta-neighborhood size /////
        is >> line;
        double max_vacant_radius;
        Conversion( max_vacant_radius, line[0] );
        
        ///// Establishment counts data /////
        is >> line;
        string location_file = line[0];
        
        ///// Area size data /////
        is >> line;
        string area_file = line[0];
        
        ///// Neighbor relations /////
        is >> line;
        string neighbors_file = line[0];
        
        ///// Border region set /////
        is >> line;
        string border_region_file = line[0];
        
        ///// Shortest-path distance /////
        is >> line;
        string distance_file = line[0];
        
        ///// Shortest-path sequence data /////
        is >> line;
        string shortest_path_file = line[0];
        
//        is >> line;
//       string shortest_path_hokkaido_file = line[0];
        
        //        ///// NML Penalty table /////
        //        is >> line;
        //        string nml_penalty_table_file = line[0];
        
        ///// Output dir /////
        is >> line;
        string out_dir = line[0];
        
        ///// Log options /////
        is >> line;
        int event_log_flag;
        Conversion(event_log_flag,line[0]);
        
        is >> line;
        int summary_log_flag;
        Conversion(summary_log_flag,line[0]);
        
        is >> line;
        int main_log_flag;
        Conversion(main_log_flag,line[0]);
        
        is >> line;
        int map_log_flag;
        Conversion(map_log_flag,line[0]);
        
        is.close();
        is.clear();
        
        ///// Read industry list /////
        
        vector<string> all_ind;
        is.open( ind_list_file.c_str() );
        
        if (!is) NoFile nf( ind_list_file );
        
        while ( is.peek() != EOF )
        {
            is >> line;
            all_ind.push_back( line[1] );
        }
        is.close();
        is.clear();
        
        ++end_id;
        
        if ( end_id > all_ind.size() )
            end_id = (int) all_ind.size();
        
        vector<string> targets;
        
        for ( int i = start_id; i < end_id; ++i )
        {
            targets.push_back( all_ind[i] );
        }
        
        
        ///// Check loaded files ////
        
        cout << "Target industries : " << targets.front() << " - " << targets.back() << endl;
//        cout << "Criterion : " <<
//        ClusterBuilder::criterionName[criterion_choice] << endl;
        cout << "Max vacant radius : " << max_vacant_radius << endl;
        cout << "Location data : " << location_file << endl;
        cout << "Area data : " << area_file << endl;
        cout << "Neighbors data : " << neighbors_file << endl;
        cout << "Border regions data : " << border_region_file << endl;
        cout << "Distance data : " << distance_file << endl;
        cout << "Shortest-path data : " << shortest_path_file << endl;
//        cout << "Shortest-path (hokkaido) data : " << shortest_path_hokkaido_file << endl;
        cout << "OUTPUT Dir: " << out_dir <<  endl;
        
        cout << "Computation log flag : " << flush;
        
        if ( bool (main_log_flag) )
            cout << "TRUE" << endl;
        else
            cout << "FALSE" << endl;
        
        cout << "Event log flag : " << flush;
        
        if ( bool (event_log_flag) )
            cout << "TRUE" << endl;
        else
            cout << "FALSE" << endl;
        
        cout << "Summary log flag : " << flush;
        
        if ( bool (summary_log_flag) )
            cout << "TRUE" << endl;
        else
            cout << "FALSE" << endl;
        
        cout << "Map log flag : " << flush;
        
        if ( bool (map_log_flag) )
            cout << "TRUE" << endl;
        else
            cout << "FALSE" << endl;
        
        
        ///// Read industry location data /////
        
        is.open( location_file.c_str() );
        
        if ( !is ) NoFile nf(location_file);
        
        IndustryLocationData data;
        data.Read(is);
        is.close();
        is.clear();
        
        cout << "\nIndustry location data read ..." << endl;
        
        
        ///// Cosntruct regions /////
        
        is.open( area_file.c_str() );
        if ( !is ) NoFile nf(area_file);
        
        set<Region*> regions; 
        // Set of all regions
        
        map<int, Region*> region_id;
        // Region ID -> Region*
        
        double total_area = 0.0;
        // National total area
        
        is >> line;
        // Header
        
        while ( is.peek() != EOF )
        {
            is >> line;
            Conversion(tmp_int,line[0]);
            Conversion(tmp_dbl,line[1]);
            
            total_area += tmp_dbl;
            Region* new_region = new Region( tmp_int, tmp_dbl );
            regions.insert( new_region );
            region_id[tmp_int] = new_region;
            
        }
        is.close();
        is.clear();
        
        for ( set<Region*>::iterator i1 = regions.begin();
             i1 != regions.end(); ++i1 )
        {
            (**i1).SetAreaShare( total_area );
        }
        
        cout << "Area data read ..." << endl;
        
        
        ///// Read neighbors data /////
        
        is.open(neighbors_file.c_str() );
        if ( !is ) NoFile nf(neighbors_file);
        
        is >> line;
        // Header
        
        while ( is.peek() != EOF )
        {
            is >> line;
            int loc; Conversion(loc, line[0]);
            int adjacent; Conversion( adjacent, line[1] );
            region_id[loc]->AddNeighbor( region_id[adjacent] );
        }
        
        is.close();
        is.clear();
        
        
        cout << "NeighborsData read ... " << endl;
        
        ///// Border regions data /////
        
        is.open( border_region_file.c_str() );
        
//        cout << "Border file opened" << endl;
        
        if ( !is ) NoFile nf(border_region_file);
        
        is >> line;
        // Header
        
        while ( is.peek() != EOF )
        {
            is >> line;
            Conversion(tmp_int,line[0]);
            
            if ( region_id.find(tmp_int) != region_id.end() )
                region_id[tmp_int]->SetBorder();
        }
        is.close();
        is.clear();
        
        cout << "\nBorder regions identified ..." << endl;
        
        
        ///// Read Distance data /////
        
        Distance distance;
        is.open( distance_file.c_str() );
        is >> line;
        // Header
        
        distance.Read(is);
        is.close();
        is.clear();
        
        cout << "Distance data read ..." << endl;
        
        ///// Read Shortest-path sequence /////
        
        ShortestPathData shortest_path_data;
        
        is.open( shortest_path_file.c_str() );
        
        is >> line;
        // Header
        
        shortest_path_data.Read(is);
        is.close();
        is.clear();
//        
//        ShortestPathData shortest_path_hokkaido_data;
//        
//        is.open( shortest_path_hokkaido_file.c_str() );
//        shortest_path_hokkaido_data.Read(is);
//        is.close();
//        is.clear();
//
        
        for ( set<Region*>::iterator i = regions.begin();
             i != regions.end(); ++i )
        {
            (**i).Read( shortest_path_data, region_id );
        }

        
//        for ( set<Region*>::iterator i = regions.begin();
//             i != regions.end(); ++i )
//        {
//            if ( (**i).Location() / 1000 > 1 )
//                (**i).Read( shortest_path_main_data,
//                           region_id );
//            else
//                (**i).Read( shortest_path_hokkaido_data,
//                           region_id );
//            
//        }
//        
        
        cout << "Shortest-path sequences data read ..." << endl;
        
        
        //        ///// Read NML table /////
        //        
        //        map<int, vector<double>* > nml_penalty;
        //        if ( criterion_choice == 2 )
        //        {
        //            is.open( nml_penalty_table_file.c_str() );
        //            
        //            while ( is.peek() != EOF )
        //            {
        //                is >> line;
        //                int estab_count; Conversion(estab_count, line[0]);
        //                nml_penalty[estab_count] = new vector<double>();
        //                
        //                for ( int i = 1; i < line.size(); ++i )
        //                {
        //                    int elm; Conversion( elm, line[i] );
        //                    nml_penalty[estab_count]->push_back(elm);
        //                }
        //            }
        //            is.close();
        //            is.clear();
        //            
        //            cout << "NML penalty table read ..." << endl;
        //            
        //        }
        //        
        //        
        
        ///// Detect clusters /////
        
        string target_ind;
        
        for ( vector<string>::iterator i0 = targets.begin();
             i0 != targets.end(); ++i0 )
        {
            cout << "Computing for Industry " << *i0 << endl;
            target_ind = *i0;
            
            ///// Customize regions to target industry /////
            
            double total_count = 0.0;
            
            for ( set<Region*>::iterator i1 = regions.begin();
                 i1 != regions.end(); ++i1 )
            {
                double new_count = 
                data.Value( (**i1).Location(), target_ind );
                
                total_count += new_count;
                (**i1).Count( new_count );
            }
            
            if ( total_count == 0.0 )
            {
                cout << "\nNo establishments in the regional system.\n" << endl;
                continue;
            }
            
            cout << "Total count = " << total_count << endl;
            cout << "Total area = " << total_area << endl;
            
            // 	  ///// Reset neighbors of each region /////
            
            // 	  for ( set<Region*>::iterator i = regions.begin();
            // 		i != regions.end(); ++i )
            // 	    {
            // 	      (**i).Read( neighbors_data, region_id );
            // 	    }
            
            ///// Setup environment for cluster-detection /////
            
            ClusterBuilder builder( //criterion_choice,
                                   &regions,
                                   max_vacant_radius,
                                   total_area,
                                   total_count,
                                   &distance
                                   //                                   ,
                                   //                                   &nml_penalty
                                   );
            
            builder.SetOS( out_dir );
            builder.SetIndustryId( target_ind );
            //	  builder.SetCriterion( criterion_choice );
            
            builder.SetLogType( ClusterBuilder::_event, bool (event_log_flag) );
            builder.SetLogType( ClusterBuilder::_summary, bool (summary_log_flag) );
            builder.SetLogType( ClusterBuilder::_main, bool (main_log_flag) );
            builder.SetLogType( ClusterBuilder::_map, bool ( map_log_flag) );
            
            if ( builder.IfTakeLog( ClusterBuilder::_event ) )
                builder.SetEventLog();
            
            if ( builder.IfTakeLog( ClusterBuilder:: _main ) )
                builder.SetCompLog();
            
            //	  cout << "ClusterBuilder set up" << endl;
            
            builder.GrowSystem();
            
        }
        
        cout << "\n\nDone!!!\n" << endl;
        
    }  
    
    catch ( const string& str )
    {
        cerr << str;
    }
    
    catch ( const char* str )
    {
        cerr << str;
    }
    
    catch ( ... )
    {
        cerr << "\nError: Something is wrong. Cannot figure out.\n";
    }
    
}
