//
//  main.cpp
//  EssentialContainment
//
//  Created by 知也 森 on 11/11/24.
//  Copyright (c) 2011年 __MyCompanyName__. All rights reserved.
//

#include <cstddef>
#include <fstream>
#include <utility>
#include <string>
#include <vector>
#include <map>
#include <set>
#include "Distance.h"
#include "Region.h"
#include "Cluster.h"
#include "IndustryLocationData.h"
#include "ShortestPathData.h"
#include "NoFile.h"
#include "CantOpenFile.h"
#include "StringManipulator.h"
#include "Csv1.h"

using namespace std;

int main()
{
    try
    {
        cout << "Finding Essential Containment ..." << endl;
        cout << "\nINPUT  : input_identify_essentialcontainment.txt"
        << endl;
        
        ofstream os, os1;
        string tmp_str;
        double tmp_dbl;
        int tmp_int;
        
        ifstream is( "./input_identify_essentialcontainment.txt" );
        Csv1 line;
        
        is >> line;
        int start_id;
        Conversion( start_id, line[0] );
        
        is >> line;
        int end_id;
        Conversion( end_id, line[0] );
        
        is >> line;
        string ind_list_file = line[0];
        
        is >> line;
        string in_best_clusters_dir = line[0];
        
        is >> line;
        string in_clusters_dir = line[0];
        
        is >> line;
        string out_dir = line[0];
        
        is >> line;
        double count_threshold;
        Conversion( count_threshold, line[0] );
        
        is >> line;
        string location_file = line[0];
        
        is >> line;
        string area_file = line[0];
        
        is >> line;
        string neighbors_file = line[0];
        
        is >> line;
        string border_region_file = line[0];
        
        is >> line;
        string distance_file = line[0];
        
        ///// Shortest-path sequence data /////
        is >> line;
        string shortest_path_main_file = line[0];
        
        is.close();
        is.clear();
        
        ///// Read industry list /////
        
        vector<string> all_ind;
        is.open( ind_list_file.c_str() );
        
        if (!is) NoFile nf( ind_list_file );
        
        while ( is.peek() != EOF )
        {
            is >> line;
            all_ind.push_back( line[1] );
        }
        is.close();
        is.clear();
        
        ++end_id;
        
        if ( end_id > all_ind.size() )
            end_id = (int) all_ind.size();
        
        vector<string> targets;
        
        for ( int i = start_id; i < end_id; ++i )
        {
            targets.push_back( all_ind[i] );
        }
        
        //////////////////////////////////////////////////
        
        cout << "Target industries : " << targets.front() << " - " << targets.back() << endl;

        cout << "Input dir for best clusters: " << in_best_clusters_dir <<  endl;
        cout << "Input dir for cluster data: " << in_clusters_dir <<  endl;
        cout << "Output dir: " << out_dir <<  endl;
        
        cout << "Cumulative BIC share : " << ( count_threshold * 100 ) << "% " << endl;
        cout << "Location data : " << location_file << endl;
        cout << "Area data : " << area_file << endl;
        cout << "Neighbors data : " << neighbors_file << endl;
        cout << "Border regions data : " << border_region_file << endl;
        cout << "Distance data : " << distance_file << endl;
        cout << "Shortest-path data : " << shortest_path_main_file << endl;
        
        ///// Read industry location data /////
        
        is.open( location_file.c_str() );
        
        if ( !is ) NoFile nf(location_file);
        
        IndustryLocationData data;
        data.Read(is);
        is.close();
        is.clear();
        
        cout << "\nIndustry location data read ..." << endl;
        
        
        ///// Read Distance data /////
        
        Distance distance;
        is.open( distance_file.c_str() );
        is >> line;
        // Header
        
        distance.Read(is);
        is.close();
        is.clear();
        
        cout << "Distance data read ..." << endl;
        
        ///// Read area data /////
        
        is.open( area_file.c_str() );
        if ( !is ) NoFile nf(area_file);
        
        set<Region*> regions; 
        // Set of all regions
        
        map<int, Region*> region_id;
        // Region ID -> Region*
        
        double total_area = 0.0;
        // National total area (contiguous)
        
        is >> line;
        // Header
        
        while ( is.peek() != EOF )
        {
            is >> line;
            Conversion(tmp_int,line[0]);
            Conversion(tmp_dbl,line[1]);
            
            total_area += tmp_dbl;
            Region* new_region = new Region( tmp_int, tmp_dbl );
            regions.insert( new_region );
            region_id[tmp_int] = new_region;
            
        }
        is.close();
        is.clear();
        
        ///// Check consistency of economic area data /////
        
        double test_area_share = 0.0;
        
        for ( set<Region*>::iterator i1 = regions.begin();
             i1 != regions.end(); ++i1 )
        {
            (**i1).SetAreaShare(total_area);
            test_area_share += (**i1).AreaShare();
        }
        
//        if ( test_area_share != 1.0 )
//            cout << "Total area share = " << test_area_share << " : Not correct." << endl;
        
        ////////////////////////////////////////////////////
        
        
        
        ///// Read neighbors data /////
        
        is.open(neighbors_file.c_str() );
        
        if ( !is ) NoFile nf(neighbors_file);
        
        is >> line;
        
        while ( is.peek() != EOF )
        {
            is >> line;
            int loc; Conversion(loc, line[0]);
            int adjacent; Conversion( adjacent, line[1] );
            region_id[loc]->AddNeighbor( region_id[adjacent] );
        }
        is.close();
        is.clear();
        
        cout << "NeighborsData read ... " << endl;
        
        ///// Border regions data /////
        
        is.open( border_region_file.c_str() );
        
        if ( !is ) NoFile nf(border_region_file);
        
        is >> line;
        
        while ( is.peek() != EOF )
        {
            is >> line;
            Conversion(tmp_int,line[0]);
            
            if ( region_id.find(tmp_int) != region_id.end() )
                region_id[tmp_int]->SetBorder();
        }
        is.close();
        is.clear();
        
        cout << "Border regions identified ..." << endl;
        
        
        
        ///// Read Shortest-path sequence /////
        
        ShortestPathData shortest_path_main_data;
        
        is.open( shortest_path_main_file.c_str() );
        
        is >> line;
        // Header
        
        
        shortest_path_main_data.Read(is);
        is.close();
        is.clear();
        
        for ( set<Region*>::iterator i = regions.begin();
             i != regions.end(); ++i )
        {
            (**i).Read( shortest_path_main_data,region_id );
        }
        
        
        cout << "Shortest-path sequences data read ..." << endl << endl;
        
        
        
        ///// Computation for each industry /////
        
        
        string outname = out_dir + "summary_essential_containment_cumulative_BIC_share_" +
        Conversion(tmp_str,(count_threshold*100), 0) +
        "_ind_" + targets[0] + "-" + targets.back() + ".csv";
        
        os1.open( outname.c_str() );
        os1 << "INDUSTRY" << ",";
        os1 << "NUM_CLUSTERS" << ",";
        os1 << "NUM_ESSENTIAL" << ",";
        os1 << "ESSENTIAL_SHARE" << ",";
        os1 << "COUNT_SHARE" << ",";
        os1 << "BIC_SHR" << ",";
        os1 << "ESSENTIAL_AREA_SHARE" << ",";
        os1 << "GLOBAL_EXTENT" << ",";
        os1 << "LOCAL_DENSITY" << endl;
        

        

        string target_ind;
        
        for ( vector<string>::iterator i0 = targets.begin();
             i0 != targets.end(); ++i0 )
        {
            cout << "Computing for Industry " << *i0 << endl;
            target_ind = *i0;
            
            ///// Customize regions to target industry /////
            
            double total_counts = 0.0;
            
            for ( set<Region*>::iterator i1 = regions.begin();
                 i1 != regions.end(); ++i1 )
            {
                double new_count = 
                data.Value( (**i1).Location(), target_ind );
                
                total_counts += new_count;
                (**i1).Count( new_count );
            }
            
            if ( total_counts == 0.0 )
            {
                cout << "\nNo establishments in the regional system.\n" << endl;
                continue;
            }
            
            
            
            ///// Determine essential clusters /////
            
            string inname = in_best_clusters_dir + "summary_ind_" + target_ind + ".csv";
            
            is.open( inname.c_str() );
            
            if (!is) NoFile nf( inname );
            
            vector<double> criterion_values;
            vector<double> counts;
            vector<int> ordered_clusters;
            
            is >> line;
            
            while ( is.peek() != EOF )
            {
                is >> line;
                
                // BIC contribution of this cluster
                double val;
                Conversion( val, line[6] );
                criterion_values.push_back( val );
                
                // Establishment counts of this cluster
                Conversion( val, line[2] );
                counts.push_back( val );
                
                // Cluster ID
                int cluster_id;
                Conversion( cluster_id, line[1] );
                ordered_clusters.push_back( cluster_id );
            }
            is.close();
            is.clear();
            
            double cluster_total_counts = 0.0;
            
            for ( vector<double>::iterator i1 = counts.begin(); 
                 i1 != counts.end(); ++i1 )
                cluster_total_counts += *i1;
            
//            cout << "Ordered cluster data read ..." << endl;
            
            
   
            
            ///// Determine inclusion threshold iteration number /////
            
            set<int> clusters;
            clusters.insert(ordered_clusters[0]);

            double included_counts = (double) counts[0];
            double criterion_value = criterion_values[0];

            for ( int i1 = 1; i1 < ordered_clusters.size(); ++i1 )
            {
                // Essential cluster conditions :
                // Cumulative BIC contribution is at least a given share of the total.
                
                if ( criterion_value < count_threshold )
                {
                    clusters.insert( ordered_clusters[i1] );
                    included_counts += counts[i1];
                    criterion_value += criterion_values[i1];
                }
                else
                    break;
            }

            
            
            
            //// Computation of essential containment /////
            
            
            inname = in_clusters_dir + "cluster_map_ind_" + target_ind + ".csv";
            
            is.open( inname.c_str() );
            if (!is) NoFile nf( inname  );
            is >> line;
            
            map<int, set<int>* > cluster_data;
            
            double essential_area_share = 0.0;
            int essential_region_counts = 0;
            
            
            while ( is.peek() != EOF )
            {
                is >> line;
                
                int center;
                Conversion(center, line[0]);
                
                int member;
                Conversion(member, line[1]);
                
                
                if ( clusters.find(center) == clusters.end() )
                {
                    // Skip if this cluster is not an essential cluster.
                    continue;
                }
                
                if ( cluster_data.find( center ) == cluster_data.end() )
                    cluster_data[center] = new set<int>;
                
                cluster_data[center]->insert(member);
                essential_area_share += region_id[member]->AreaShare();
                ++essential_region_counts;

            }
            is.close();
            is.clear();
            
            
//            cout << "\nCluster scheme read ..." << endl;
            
            
            
            set<Cluster*> ec;
            // Disconnected components of the essential clusters are identified separately,
            // and stored in ec.

            set<int> clusters_left;
            
            for (set<int>::iterator i1 = clusters.begin(); i1 != clusters.end(); ++i1)
                clusters_left.insert(*i1);

            Cluster* target_cluster;
            Cluster::total_area = total_area;
            
            while ( clusters_left.size() > 0 )
            {
                target_cluster = new Cluster( region_id[*clusters_left.begin()], &distance);
                set<int> added_clusters;
                added_clusters.insert(target_cluster->Location());
                
                for (set<int>::iterator i1 = clusters_left.begin(); i1 != clusters_left.end(); ++i1 )
                {
//                    if ( *i1 == target_cluster->Location() || distance(target_cluster->Location(), *i1) == DBL_MAX )
                    if ( distance(target_cluster->Location(), *i1) == DBL_MAX )
                        continue;
                    
                    for ( set<int>::iterator i2 = cluster_data[*i1]->begin();
                         i2 != cluster_data[*i1]->end(); ++i2 )
                    {
                        target_cluster->AddCore(region_id[*i2]);
                        target_cluster->AddDConvexSolid(region_id[*i2]);
                    }
                    
                    added_clusters.insert(*i1);
                }
                
                if ( added_clusters.size() > 1 )
                {
                    // If clusters to be convex-solidfied consists of only one cluster, the areal size
                    // could exceed the original cluster, since essential containment D-convexifies
                    // the entire set of member basic regions.
                    
                    target_cluster->DConvexSolidifyFull(regions);
                }
                
                target_cluster->CalcCountAndArea();
                target_cluster->CalcAreaShare();
                
                ec.insert(target_cluster);
                
                for (set<int>::iterator i2 = added_clusters.begin(); i2 != added_clusters.end(); ++i2 )
                    clusters_left.erase(*i2);
                
            }
            
            ///// Output for this industry /////
            
            double essential_containment_area = 0.0;
            double member_area  = 0.0;
            int member_count = 0;
            int containment_counts = 0;
            
            outname = out_dir + "essential_containment_cumulative_BIC_share_" +
            Conversion(tmp_str, (count_threshold*100),0) +
            "_ind_" + target_ind + ".csv";
            
            os.open( outname.c_str() );
            os << "BASE, MEMBER" << endl;
            
            cout << endl;
            
            for (set<Cluster*>::iterator i1 = ec.begin(); i1 != ec.end(); ++i1 )
            {
                essential_containment_area += (**i1).AreaShare();
                containment_counts += (**i1).Count();
                cout << "Component " << (**i1).Location() << "'s area share = " << (**i1).AreaShare() << endl;
                cout << "Component " << (**i1).Location() << "'s region counts = " << (**i1).DConvexSolid().size() << endl;
                
                for ( set<Region*>::iterator i2 = (**i1).DConvexSolid().begin();
                     i2 != (**i1).DConvexSolid().end(); ++i2 )
                {
                    os << (**i1).Location()<< ", " << (**i2).Location() << endl;
                    member_area += (**i2).AreaShare();
                    ++member_count;
                }
            }
            
            cout << "\nArea share (e-containment) = " << member_area << endl;
            cout << "Region counts (e-containment) = " << member_count << endl;
            cout << "Area share (essential clusters) = " << essential_area_share << endl;
            cout << "Region counts (essential regions) = " << essential_region_counts << endl << endl;
            
          
            os.close();
            os.clear();
            
            
            ///// Summary output /////
           
            os1 << *i0 << ",";
            os1 <<(int) ordered_clusters.size() << ",";
            os1 << clusters.size() << ",";
            os1 << ( (double) clusters.size() / (double) ordered_clusters.size() ) << ",";
            os1 <<  ( included_counts / cluster_total_counts ) << ",";
            os1 << criterion_value << ",";
            os1 << essential_area_share << ",";
            os1 << essential_containment_area << ",";
            os1 << (essential_area_share/essential_containment_area) << endl;
            
            for (set<Cluster*>::iterator i1 = ec.begin(); i1 != ec.end(); ++i1 )
                delete *i1;
            
//            cout << "Establishment counts in essential clusters : "<< essential_counts << endl;
//            cout << "Establishment counts in e-containment : " << containment_counts;
        }

        os1.close();
        os1.clear();


        cout << "\n\nDone!!!\n" << endl;
    }  
    
    catch ( const string& str )
    {
        cerr << str;
    }
    catch ( const char* str )
    {
        cerr << str;
    }
    catch ( ... )
    {
        cerr << "\nError: Something is wrong. Cannot figure out.\n";
    }
}
