#
# FindAgglomerations.py
#

targets = ['127','129','151','173','252','284','308']

country = 'jp'
year = '2000'
basedir = '../DATA/jp/'
industry_file = basedir + 'industry_list_' + country + '_' + year + '.csv'
border_file = basedir + 'boundaries_' + country + '_' + year + '.csv'
#biwako_border_file = basedir + 'biwa_lake_borders.csv'
region_file = basedir + 'municipality_economic-area_' + country + '_' + year + '.csv'
neighbor_file = basedir + 'neighbors_' + country + '_' + year + '.csv'
cluster_head = basedir + 'clusters/members/cluster_map_ind_'
size_head = basedir + 'clusters/sizes/cluster_estab_size_'
#emp_head =  basedir + 'clusters/size/cluster_emp_size_'
bic_head = basedir + 'clusters/summary/summary_ind_'
out_member_head = basedir + 'agglomerations/members/agglomeration_map_'
out_cluster_head = basedir + 'agglomerations/clusters/agglomeration_clusters_'
out_size_head = basedir + 'agglomerations/sizes/agglomeration_size_'

import csv, sys

### Industries ###
infile = open(industry_file,'r')
reader = csv.reader(infile)
industries = []
for row in reader:
    industries.append(str(row[0]))

infile.close

### Basic regions ###
infile = open(region_file,'r')
reader = csv.reader(infile)

regions = set()
area = {}

for row in reader:
    region = int(row[0])
    value = float(row[1])
    regions.add(region)
    area[region] = value
    
infile.close

### Neighbors ###
infile = open(neighbor_file,'r')
reader = csv.reader(infile)
reader.next()
neighbors = {}
for row in reader:
    [center,neighbor] = map(int,row)
    
    if ( center in neighbors ) == False:
	  neighbors[center] = set()
	  
    neighbors[center].add(neighbor)

infile.close

### Borders ###
infile = open(border_file,'r')
reader = csv.reader(infile)
reader.next()
borders = []
for row in reader:
    borders.append(int(row[0]))
    
infile.close
#
#infile = open(biwako_border_file,'r')
#reader = csv.reader(infile)
#for row in reader:
#    borders.append(int(row[0]))
#
#infile.close

### Agglomerations ###
if len(targets) == 0:
    targets = industries
    
for industry in targets:
 
    if ( industry in industries ) == False:
	  sys.stdout.write('Skipping industry ')
	  sys.stdout.write(industry)
	  sys.stdout.flush()
	  continue
    
    sys.stdout.write('Computing for industry ' + industry + '...\n')
    sys.stdout.flush()

    ### Identify clusters ###
    inname = cluster_head + industry + '.csv'
    infile = open(inname,'r')
    reader = csv.reader(infile)
    reader.next()
    
    clusters = {}
    cluster_area = {}

    for row in reader:
	  [center,member] = map(int,row)

	  if ( center in clusters ) == False:
		clusters[center] = set()
		cluster_area[center] = 0.0
		
	  clusters[center].add(member)
	  cluster_area[center] += area[member]

    infile.close
    
    ### Establishment number ###
    inname = size_head + industry + '.csv'
    infile = open(inname,'r')
    reader = csv.reader(infile)
    reader.next()
    cluster_size = {}
    cluster_density = {}
    
    for row in reader:
	  center = int(row[0])
	  value = float(row[1])

	  cluster_size[center] = value
	  cluster_density[center] = value/cluster_area[center]
    
    infile.close
    
   
    ### BIC contribution ###
    inname = bic_head + industry + '.csv'
    infile = open(inname,'r')
    reader = csv.reader(infile)
    reader.next()
    bic = {}
    for row in reader:
	  cluster = int(row[1])
	  bic[cluster] = float(row[-1])
#	  print [cluster, bic[cluster]]
	  
    infile.close
	  
    ### Neighbor clusters ###
    cluster_neighbors = {}
    
#    print 'Neighbors : '
    for i in clusters:
	  cluster_neighbors[i] = []
		
	  for j in clusters:
	      if i == j:
		    continue
		    
	      for member in clusters[i]:
		    if neighbors[member].intersection(clusters[j]):
			  cluster_neighbors[i].append(j)
			  break
		    
#	  print [i, cluster_neighbors[i]]

    ### Peak clusters ###
    
    def bic_compare(x):
	  return -bic[x]

    def density_compare(x):
	  return -cluster_density[x]
    
    peak = {}
    
    for i in clusters:
	  if len(cluster_neighbors[i]) == 0:
		peak[i] = i
		continue
		
	  cluster_neighbors[i].sort(key=density_compare)
	  largest_density_neighbor = cluster_neighbors[i][0]
	  if cluster_density[i] < cluster_density[largest_density_neighbor]:
		peak[i] = largest_density_neighbor
	  else:
		peak[i] = i

#    print 'Peaks : '
#    print len(peak)

#    for i in peak:
#	  print [i, peak[i]]
	  
    ### Agglomerations ###
    def global_peak(x):
	  if peak[x] == x:
		return x
	  else:
		return global_peak(peak[x])

    agglomerations = {}
    agglomeration_bic = {}
    agglomeration_clusters = {}

	
#    print 'Global peaks: '
    for i in clusters:
	  center = global_peak(i)

#	  print [i, center]
	  if (center in agglomerations) == False:
		agglomerations[center] = set()
		agglomeration_clusters[center] = []
		agglomeration_bic[center] = 0.0
		
		
	  agglomeration_bic[center] += bic[i]
	  agglomerations[center] = agglomerations[center].union(clusters[i])
	  agglomeration_clusters[center].append(i)

#    print 'Agglomerations : '
#    for i in agglomerations:
#	  print [i, len(agglomerations[i])]


#    region_agglomeration = {}
#    for i in agglomerations:
#	  for r in agglomerations[i]:
#		region_agglomeration[r] = i




### Fill in holes in agglomerations ###
    interior_complements ={}
    
    
    
    for i in agglomerations:
	  
	  ### If agglomeration consists of only two regions,
	  ### there should be no hole.
	  
	  if len(agglomerations[i]) < 3:
		interior_complements[i] = set([])
		continue
	  
	  ### Regions outside agglomeration i ###
	  agglomeration_complement = regions - agglomerations[i]
	  

	  def exteriorset(x, result):
		### region x should be in the complement set of agglomeration ###
		### Neighbors of x outside agglomeration i is recursively
		### included in "result" set.
				
		for region in neighbors[x]:
		    
		    if (region in agglomerations[i]) or (region in result):
			  continue

		    result.add(region)
		    exteriorset(region,result)
		
        ### Identify exterior complement of i-th agglomeration ###
	  exterior_complement = set()
	  
	  for region in borders:
		### Start from each border region ###
		
		if region in agglomeration_complement:
		    ### if a border region is in complement of agglomeration,
		    ### it is an exterior complement of the agglomeration.
		    exterior_complement.add(region)
		    
		    ### Identify all regions of exterior complement of agglomeration i
		    exteriorset(region, exterior_complement)

	  ### Interior complement of agglomeration i ###
	  ### (It may contain other agglomerations of the same industry.)
	  interior_complements[i] = agglomeration_complement - exterior_complement
	  
	  
	  
    ### Check if interior complement of each agglomeration contains other agglomerations ###
    agglomeration_set = agglomerations.keys()
    agglomeration_set.sort()
    
    delete_agglomerations = set()
    
#    for i in range(0,len(agglomeration_set)):
    for i in agglomeration_set:
	  if i in delete_agglomerations:
		continue
	  
#	  for j in range(0,len(agglomeration_set)):
	  for j in agglomeration_set:
		if (j == i) or (j in delete_agglomerations):
		    continue
#		sys.stdout.write(str(i))
#		sys.stdout.write(' : ')
		#sys.stdout.write(str(j))
		#sys.stdout.write('\n')
		#sys.stdout.flush()
		if len(interior_complements[i].intersection(agglomerations[j])) > 0:

		    ### Interior complement of agglomeration i contains agglomeration j ###
		    if cluster_density[i] > cluster_density[j]:
			  
			  ### Center of agglomeration i has higher establishment density than that of j ###
			  sys.stdout.write('\nConsolidating ')
			  sys.stdout.write(str(j))
			  sys.stdout.write(' into ')
			  sys.stdout.write(str(i))
			  sys.stdout.write('\n')
			  delete_agglomerations.add(j)
			  agglomerations[i] = agglomerations[i].union(agglomerations[j])
			  agglomeration_clusters[i].extend(agglomeration_clusters[j])
			  agglomeration_bic[i] += agglomeration_bic[j]
			  interior_complements[i] = interior_complements[i] - agglomerations[j]
			  interior_complements[i] = interior_complements[i].union(interior_complements[j])
		    else:
			  ### Center of agglomeration j has higher establishment density than that of i ###
  			  sys.stdout.write('Consolidating ')
			  sys.stdout.write(str(i))
			  sys.stdout.write(' into ')
			  sys.stdout.write(str(j))
			  sys.stdout.write('\n')

			  delete_agglomerations.add(i)
			  agglomerations[j] = agglomerations[j].union(agglomerations[i])
			  agglomeration_clusters[j].extend(agglomeration_clusters[i])
			  agglomeration_bic[j] += agglomeration_bic[i]
			  interior_complements[j] = interior_complements[i] - agglomerations[j]
			  interior_complements[j] = interior_complements[i].union(interior_complements[j])
			  
			  
    		  
    for i in delete_agglomerations:
	  del agglomerations[i]
	  del agglomeration_clusters[i]
	  del agglomeration_bic[i]
	  
    ####

    for i in agglomerations:
	  agglomerations[i] = agglomerations[i].union(interior_complements[i])
    
    
    
    
    
	  
	  
	  
		
    ### Output ###
    ## Agglomeration members ##    
    outname = out_member_head + industry + '_' + country + '_' + year + '.csv'
    outfile = open(outname,'w')
    writer = csv.writer(outfile)
    writer.writerow(['CENTER', 'MEMBER'])
    centers = agglomerations.keys()
    centers.sort()
    
    for i in centers:
	  members = list(agglomerations[i])
	  members.sort()
		
	  for j in members:
		writer.writerow([i, j])

    outfile.close
    
    ### Consolidated clusters ###
    outname = out_cluster_head + industry + '_' + country + '_' + year + '.csv'
    outfile = open(outname,'w')
    writer = csv.writer(outfile)
    writer.writerow(['CENTER', 'CLUSTER'])
    
    for i in centers:
	  agglomeration_clusters[i].sort()
	  
	  for j in agglomeration_clusters[i]:
	  	writer.writerow([i,j])
	  
    outfile.close	  
	  

    ### Agglomeration BIC contributions and size ###
    outname = out_size_head + industry + '_' + country + '_' + year + '.csv'
    outfile = open(outname,'w')
    writer = csv.writer(outfile)
    writer.writerow(['CENTER', 'NUM_CLUSTERS', 'BIC', 'SIZE'])
    
    for i in centers:
	  total_size = 0
	  for j in agglomeration_clusters[i]:
		total_size += cluster_size[j]
		
	  writer.writerow([i, len(agglomeration_clusters[i]), agglomeration_bic[i], total_size])
    
    outfile.close
	  
    
   
    
    
    
    